#include <common.h>

#ifdef CONFIG_SERIAL_FLASH

#include <serial_flash.h>

serial_flash_info_t sflash_info;

static void serial_flash_update_status (serial_flash_info_t *);
static int  get_real_addr (ulong, ulong *);

/************************************************************************
 * Function name:	serial_flash_init

 * Description	:	

 * Input value	:	None	
 * Return value	:	size of serial flash
 ***********************************************************************/
int serial_flash_init (void)
{
	unsigned int i;
	ulong temp;

	SF_init();

	memset (&sflash_info, 0, sizeof(serial_flash_info_t));

	SF_read_id (&temp);
	switch (temp)
	{
	case M25P64:
		sflash_info.serial_flash_id = M25P64;
		sflash_info.size = 0x800000;
		sflash_info.sector_size = 0x10000;
		sflash_info.sector_count = 128;
		sflash_info.page_size = 256;
		break;
	case M25P16:
		sflash_info.serial_flash_id = M25P16;
		sflash_info.size = 0x200000;
		sflash_info.sector_size = 0x10000;
		sflash_info.sector_count = 32;
		sflash_info.page_size = 256;
		break;
	default:
		sflash_info.serial_flash_id = SERIAL_FLASH_UNKNOWN;
		sflash_info.size = 0;
		sflash_info.page_size = 0;
		sflash_info.sector_count = 0;
		return SFLASH_UNKNOWN_FLASH;
	}

	sflash_info.logical_address = CFG_SFLASH_LOGIC_ADDR;
	
	for (i = 0; i < sflash_info.sector_count; i++) {
		sflash_info.sector_protect[i] = NOT_PROTECTED;
		sflash_info.sector_start[i] = i * sflash_info.sector_size + sflash_info.logical_address;
	}

	serial_flash_update_status (&sflash_info);
	
	return sflash_info.size;
}


/************************************************************************
 * Function name:	serial_flash_update_status

 * Description	:	When the protect bits in the status register are
			adjusted, this function is called to update the
			sector_protect array.

 * Input value	:	pointer to serial flash info struct
 * Return value	:	0 on success, 1 on failure
 ***********************************************************************/

static void serial_flash_update_status (serial_flash_info_t *info)
{
	unsigned int i;
	
	if (info->serial_flash_id == M25P64)
	{
		uchar sr;
		SF_read_status_register (&sr);
		
		switch ((sr & SERIAL_FLASH_SR_BP_MASK) >> 2)
		{
		case 0:
			for (i = 0; i < info->sector_count; i++)
				info->sector_real_protect[i] = NOT_PROTECTED;
			break;
		case 1:
			for (i = 0; i < 126; i++)
				info->sector_real_protect[i] = NOT_PROTECTED;
			for (i = 126; i < info->sector_count; i++)
				info->sector_real_protect[i] = PROTECTED;
			break;
		case 2:
			for(i = 0; i < 124; i++)
				info->sector_real_protect[i] = NOT_PROTECTED;
			for(i = 124; i < info->sector_count; i++)
				info->sector_real_protect[i] = PROTECTED;
			break;
		case 3:
			for(i = 0; i < 120; i++)
				info->sector_real_protect[i] = NOT_PROTECTED;
			for(i = 120; i < info->sector_count; i++)
				info->sector_real_protect[i] = PROTECTED;
			break;
		case 4:
			for(i = 0; i < 112; i++)
				info->sector_real_protect[i] = NOT_PROTECTED;
			for(i = 112; i < info->sector_count; i++)
				info->sector_real_protect[i] = PROTECTED;
			break;
		case 5:
			for(i = 0; i < 96; i++)
				info->sector_real_protect[i] = NOT_PROTECTED;
			for(i = 96; i < info->sector_count; i++)
				info->sector_real_protect[i] = PROTECTED;
			break;
		case 6:
			for(i = 0; i < 64; i++)
				info->sector_real_protect[i] = NOT_PROTECTED;
			for(i = 64; i < info->sector_count; i++)
				info->sector_real_protect[i] = PROTECTED;
			break;
		case 7:
			for(i = 0; i < info->sector_count; i++)
				info->sector_real_protect[i] = PROTECTED;
			break;
		default:
			return;
		}
	}
	return;	
}



/************************************************************************
 * Function name:	get_real_addr

 * Description	:	Checks if the given logical address maps to
			a real serial flash address and finds the
			'real' address

 * Input value	:	logical address, ptr to the real addr value
 * Return value	:	1 if the address is valid, 0 if it is not
 ***********************************************************************/
int get_real_addr (ulong logic_addr, ulong *real_addr)
{
	ulong temp_addr;
	if ((logic_addr & CFG_SFLASH_LOGIC_ADDR_MASK) == sflash_info.logical_address) {
		temp_addr = logic_addr - sflash_info.logical_address;
		if (temp_addr >= sflash_info.size)
			return 0;
		*real_addr = temp_addr;
		return 1;
	}
	return 0;
}


/************************************************************************
 * Function name:	serial_flash_addr

 * Description	:	External function to check if the given logical 
			address maps to	serial flash memory

 * Input value	:	address
 * Return value	:	1 if the address is valid, 0 if it is not
 ***********************************************************************/
int serial_flash_addr (ulong addr)
{
	if ((addr & CFG_SFLASH_LOGIC_ADDR_MASK) == sflash_info.logical_address) {
		if ((addr - sflash_info.logical_address) < sflash_info.size)
			return 1;
		return 0;
	}
	return 0;
}




/************************************************************************
 * Function name:	serial_flash_print_info

 * Description	:	Prints the serial flash manufacturer, the chip
			type, size, sector count, page size, the
			start addresses of sectors and whether the
			latter are read only or not

 * Input value	:	None
 * Return value	:	0 on success, 1 on failure
 ***********************************************************************/
void serial_flash_print_info (void)
{
	unsigned int i;

	switch (sflash_info.serial_flash_id)
	{
	case M25P64:
		puts ("STMicroelectronics ");
		puts ("M25P64 (64Mbit, uniform sectors\n");
		break;
	default:
		puts ("Unknown serial flash\n");
		return;
	}
	
	
	printf ("   Size:\t%lu MB in %u sectors\n", 
		sflash_info.size >> 20,
		sflash_info.sector_count);
	printf ("   Page size:\t%u bytes\n", sflash_info.page_size);

	printf ("   Logical address:\t%08lX\n", sflash_info.logical_address);

	puts ("   Sector start addresses:");
	
	for (i = 0; i < sflash_info.sector_count; i++)
	{
		if ((i % 5) == 0)
			puts ("\n   ");

		printf (" %08lX%s", sflash_info.sector_start[i], 
			(sflash_info.sector_protect[i] || sflash_info.sector_real_protect[i]) ? " (RO)" : "     ");
	}
	puts ("\n");	
	return;
}



/************************************************************************
 * Function name:	serial_flash_read

 * Description	:	Reads 'size' bytes from flash memory starting at
			address 'addr' and stores the outcome in result.
			As it is now, no loop reading is supported, i.e. 
			it will read up to the end address of memory 
			and return.

 * Input value	:	address, size to be read, and a pointer to
			where to store the outcome 
 * Return value	:	0 on success, 1 on failure
 ***********************************************************************/
int serial_flash_read (ulong logic_addr, ulong size, uchar *result)
{
	ulong addr;
	unsigned int exceeds_limit = 0;

	/* Check if the supplied address is valid */
	if (!(get_real_addr(logic_addr, &addr)))
		return SFLASH_INVAL;
	
	/* Check if the address and size combo is valid, i.e. adds up to
	 * a location inside memory
	 *
	 * NOTE: The M25P64 does allow indefinite loop reading of memory,
	 * however, this function does not allow it!
	 */
	if ((addr + size) > sflash_info.size)
	{
		puts ("\n- Warning: Read exceeds memory limits\n");
		exceeds_limit = 1;		
	}
	
	if (exceeds_limit)
		SF_fast_read (addr, sflash_info.size - addr, result);
	else
		SF_fast_read (addr, size, result);

	return SFLASH_OK;
}


/************************************************************************
 * Function name:	serial_flash_write

 * Description	:	Writes 'size' bytes from 'data' to flash memory 
			starting at address 'addr'.

 * Input value	:	address, size to be written, and a pointer to
			where to read the data from
 * Return value	:	0 on success, 1 on failure
 ***********************************************************************/
int serial_flash_write (ulong logic_dest_addr, ulong size, ulong source_addr)
{
	unsigned int i;
	ulong dest_addr;
	ushort start_sector;	/* Starting sector to be programmed */
	ushort end_sector;		/* Last sector to be programmed */

	/* Page Programming variables */
	ulong from_addr;
	ulong to_addr;
	ulong to_program_size;
	ulong remaining_pages;

	/* Input Parameter Checking */
	/****************************/
	/* Check if the supplied address is valid */
	if (!get_real_addr(logic_dest_addr, &dest_addr))
		return SFLASH_INVAL;
	
	/* Check if the address and size combo is valid, i.e. adds up to
	 * a location inside memory
	 */
	if ((dest_addr + size) > sflash_info.size)
		return SFLASH_OVERFLOW;

	/* Check if the area to be written to is protected or not */
	start_sector = (dest_addr / sflash_info.sector_size);
	end_sector = (dest_addr + size - 1) / sflash_info.sector_size;

	for (i = start_sector; i <= end_sector; i++) {
		if (sflash_info.sector_protect[i] == PROTECTED || sflash_info.sector_real_protect[i] == PROTECTED)
			return SFLASH_PROTECTED;
	}

	/* Write the data */
	/*****************/
	/* The very first page might not be a whole one if the destination address does not
	   end with 00H. -> Program starting at destination address to the end of the current
	   page. */
	
	to_program_size = sflash_info.page_size - (dest_addr % sflash_info.page_size); 
	
	if (SF_page_program(dest_addr, to_program_size, (char *) source_addr))
		return SFLASH_TIMOUT;
	
	from_addr = source_addr + to_program_size;
	to_addr = dest_addr + to_program_size;
	
	/* If the last page to be programmed is partial, it is not accounted for in remaining_pages */
	remaining_pages = ((size - to_program_size) / sflash_info.page_size);
	
	for (i = 0; i < remaining_pages; i++)
	{
		if (SF_page_program(to_addr, sflash_info.page_size, (char *) from_addr))
			return SFLASH_TIMOUT;

		from_addr += sflash_info.page_size;
		to_addr += sflash_info.page_size;
	}
	
	/* Check if there is one last page to be partially programmed */
	/* The last address is dest_addr + size - 1 */
	if (((dest_addr + size) & 0xFF) != 0x00)
	{
		to_program_size = (dest_addr + size - to_addr);
		if (SF_page_program(to_addr, to_program_size, (char *) from_addr))
			return SFLASH_TIMOUT;
	}

	return SFLASH_OK;
}


/************************************************************************
 * Function name:	serial_flash_erase

 * Description	:	Erases unprotected sectors within a range
			defined by sector numbers 

 * Input value	:	start and end sectors' numbers
 * Return value	:	0 on success, 1 on failure
 ***********************************************************************/
int serial_flash_erase (ushort start_sector, ushort end_sector)
{
	unsigned int i;
	ushort erased_count;
	ushort protected_count;

	if ((start_sector >= sflash_info.sector_count) ||
	   (end_sector >= sflash_info.sector_count) ||
	   (start_sector > end_sector))
		return SFLASH_INVAL;
	
	erased_count = 0;
	protected_count = 0;

	/* Issue warning that protected sector will not be erased */
	for (i = start_sector; i <= end_sector; i++)
	{
		if (sflash_info.sector_protect[i] == PROTECTED || sflash_info.sector_real_protect[i] == PROTECTED)
			protected_count++;
	}

	if (protected_count > 0)
		printf ("\n- Warning: %hu protected sectors will not be erased!\n", protected_count);
	
	for (i = start_sector; i <= end_sector; i++)
	{
		/* Check if there are protected sectors within the given range */
		if (sflash_info.sector_protect[i] == PROTECTED || sflash_info.sector_real_protect[i] == PROTECTED)
			continue;
		
		if (SF_sector_erase(sflash_info.sector_start[i] - sflash_info.logical_address)) {
			printf("Erased %hu sectors.\n", erased_count);
			return SFLASH_TIMOUT;
		}
		erased_count++;
	}
	puts (" done\n");

	if (erased_count > 0)
		printf ("Erased %hu sectors.\n", erased_count);
	else
	{
		puts ("No sectors were erased.\n");
	}

	return SFLASH_OK;
}


/************************************************************************
 * Function name:	serial_flash_sect_erase

 * Description	:	Erases unprotected sectors within a range
			defined by memory addresses

 * Input value	:	start and end sectors' addresses
 * Return value	:	0 on success, 1 on failure
 ***********************************************************************/
int serial_flash_sect_erase (ulong logic_start_addr, ulong logic_end_addr)
{
	unsigned int i;
	ushort erased_count;
	ushort protected_count;
	ushort s_start, s_end;
	ulong start_addr, end_addr;

	if (!get_real_addr(logic_start_addr, &start_addr) ||
	    !get_real_addr(logic_end_addr, &end_addr) ||
	    (logic_start_addr > logic_end_addr))
		return SFLASH_INVAL;
	
	erased_count = 0;
	protected_count = 0;

	/* SF_get_set_number is defined as a Macro */	
	s_start = SF_get_sect_number(start_addr);
	s_end = SF_get_sect_number(end_addr);

	/* Issue warning that protected sector will not be erased */
	for (i = s_start; i <= s_end; i++)
	{
		if (sflash_info.sector_protect[i] == PROTECTED || sflash_info.sector_real_protect[i] == PROTECTED)
			protected_count++;
	}

	if (protected_count > 0)
		printf("\n- Warning: %hu protected sectors will not be erased!\n", protected_count);
	
	for (i = s_start; i <= s_end; i++)
	{
		/* Check if the current sector to be erased is protected */
		if (sflash_info.sector_protect[i] == PROTECTED || sflash_info.sector_real_protect[i] == PROTECTED)
			continue;	
		if (SF_sector_erase(sflash_info.sector_start[i] - sflash_info.logical_address)) {
			printf("Erased %hu sectors.\n", erased_count);
			return SFLASH_TIMOUT;
		}
		erased_count++;
	}
	puts (" done\n");


	if (erased_count > 0)
		printf ("Erased %hu sectors.\n", erased_count);
	else
	{
		puts ("No sectors were erased.\n");
	}
	return SFLASH_OK;
}



/************************************************************************
 * Function name:	serial_flash_bulk_erase

 * Description	:	Erases all unprotected sectors. 

 * Input value	:	start and end sectors
 * Return value	:	0 on success, 1 on failure
 ***********************************************************************/
/* NOTE TO SELF: BULK ERASE IS NOT PERFORMED IF ONE OR MORE SECTORS ARE 
   PROTECTED.  IS THERE A NEED FOR A SOFTWARE BULK ERASE? */
int serial_flash_bulk_erase (void)
{
	unsigned int i;
	ushort protected_count;
	ushort erased_count;

	protected_count = 0;
	erased_count = 0;

	/* Check if there are protected sectors */
	for (i = 0; i < sflash_info.sector_count; i++)
	{
		if (sflash_info.sector_protect[i] == PROTECTED || sflash_info.sector_real_protect[i] == PROTECTED)
			protected_count++;
	}
	
	if (protected_count == sflash_info.sector_count)
		return SFLASH_ALL_PROT;

	if (protected_count > 0) {
		printf("\n- Warning: %hu protected sectors will not be erased!\n", protected_count);
		for (i = 0; i < sflash_info.sector_count; i++) {
			if (sflash_info.sector_protect[i] == PROTECTED || sflash_info.sector_real_protect[i] == PROTECTED)
				continue;
			if (SF_sector_erase(sflash_info.sector_start[i] - sflash_info.logical_address)) {
				printf("\nErased %hu sectors\n.", erased_count);
				return SFLASH_TIMOUT;
			}
			erased_count++;
		}
		puts (" done\n");
		printf("Erased %hu sectors\n", erased_count);
		return SFLASH_OK;
	}

	if (SF_bulk_erase())
		return SFLASH_TIMOUT;

	puts (" done\n");
	puts ("Erased all sectors.\n");

	return SFLASH_OK;
}



/************************************************************************
 * Function name:	serial_flash_block_protect

 * Description	:	REAL-protects a block of sectors according to
			the value specified by bp

 * Input value	:	char bp
 * Return value	:	0 on success, 1 on failure
 ***********************************************************************/
int serial_flash_block_protect (uchar bp)
{
	if(bp > SERIAL_FLASH_SR_BP_MAX)
	{
		printf ("Block protect input %c is not valid! Aborting protect.\n", bp);
		return SFLASH_INVAL;
	}
	
	SF_block_protect (bp);

	serial_flash_update_status (&sflash_info);

	return SFLASH_OK;
}



/************************************************************************
 * Function name:	serial_flash_protect

 * Description	:	Turns write-protection on/off for a range of
			sectors defined by start and end addresses

 * Input value	:	on/off flag, from address, to address
 * Return value	:	0 on success, 1 on failure
 ***********************************************************************/
int serial_flash_protect (int flag, ulong logic_start_addr, ulong logic_end_addr)
{
	int i;
	ulong start_addr, end_addr;
	ushort s_start, s_end;

	if (!get_real_addr(logic_start_addr, &start_addr) ||
	    !get_real_addr(logic_end_addr, &end_addr) ||
	    (start_addr > end_addr))
		return ERR_INVAL;

	s_start = SF_get_sect_number(start_addr);
	s_end = SF_get_sect_number(end_addr);

	for (i = s_start; i <= s_end; i++)
		sflash_info.sector_protect[i] = flag ? PROTECTED : NOT_PROTECTED;

	return SFLASH_OK;
}

/************************************************************************
 * Function name:	serial_flash_perror

 * Description	:	

 * Input value	:	error code	
 * Return value	:	none
 ***********************************************************************/
void serial_flash_perror (int err)
{
	puts("\n");

	switch (err) {
	case SFLASH_OK:
		break;
	case SFLASH_TIMOUT:
		puts ("Serial flash timeout\n");
		break;
	case SFLASH_PROTECTED:
		puts ("Can't write to protected serial flash sectors\n");
		break;
	case SFLASH_INVAL:
		puts ("Invalid serial flash memory specification\n");
		break;
	case SFLASH_UNKNOWN_FLASH:
		puts ("Unknown type of serial flash\n");
		break;
	case SFLASH_OVERFLOW:
		puts ("Read/Write exceeds serial flash memory limits\n");
		break;
	case SFLASH_ALL_PROT:
		puts ("Entire serial flash is protected\n");
		break;
	default:
		printf ("%s[%d] FIXME: rc=%d\n", __FILE__, __LINE__, err);
		break;
	}
}

#endif /* CONFIG_SERIAL_FLASH */

